import { setQuerybuilderNodes, store } from '@graphpolaris/shared/lib/data-access/store';
import Graph, { MultiGraph } from 'graphology';
import { Attributes as GAttributes, Attributes, SerializedGraph } from 'graphology-types';
import { EntityNodeAttributes, LogicNodeAttributes, QueryGraphNodes, RelationNodeAttributes } from './model';
import { XYPosition } from 'reactflow';
import { Handles, QueryElementTypes } from '../reactflow';
import { getHandleId } from '..';

/** monospace fontsize table */
const widthPerFontsize = {
  6: 3.6167,
  7: 4.2167,
  10: 6.0167,
};

export type QueryMultiGraph = SerializedGraph<QueryGraphNodes, GAttributes, GAttributes>;

export class QueryMultiGraphology extends MultiGraph<QueryGraphNodes, GAttributes, GAttributes> {
  public addPill2Graphology(attributes: QueryGraphNodes): QueryGraphNodes {
    const { type, name } = attributes;
    if (!type || !name) throw Error('type or name is not defined');
    if (!attributes.id) attributes.id = 'id_' + (Date.now() + Math.floor(Math.random() * 1000)).toString();
    let { x, y } = attributes;

    // Check if x and y are present, otherwise set them to 0
    if (!x) x = 0;
    if (!y) y = 0;

    // Get the width and height of a node
    const { width, height } = calcWidthHeightOfPill(attributes);

    // fix handles
    if (type === QueryElementTypes.Entity) {
      (attributes as EntityNodeAttributes).leftRelationHandleId = getHandleId(attributes.id, name, type, Handles.EntityLeft, '');
      (attributes as EntityNodeAttributes).rightRelationHandleId = getHandleId(attributes.id, name, type, Handles.EntityRight, '');
    } else if (type === QueryElementTypes.Relation) {
      (attributes as RelationNodeAttributes).leftEntityHandleId = getHandleId(attributes.id, name, type, Handles.RelationLeft, '');
      (attributes as RelationNodeAttributes).rightEntityHandleId = getHandleId(attributes.id, name, type, Handles.RelationRight, '');
    } else if (type === QueryElementTypes.Logic) {
    }

    // Add a node to the graphology object
    const nodeId = this.addNode(attributes.id, { ...attributes, x, y, width, height });

    // Set the new nodes in the query builder slice TODO: maybe remove for efficiency
    store.dispatch(setQuerybuilderNodes(this.export()));

    return attributes;
  }

  public addEdge2Graphology(source: QueryGraphNodes, target: QueryGraphNodes, attributes: GAttributes): string | null {
    // fix handles
    attributes.sourceHandle = getHandleId(
      source.id + '_TO_' + target.id,
      source.name,
      source.type,
      source.type === QueryElementTypes.Entity ? Handles.EntityRight : Handles.RelationRight,
      ''
    );
    attributes.targetHandle = getHandleId(
      source.id + '_TO_' + target.id,
      target.name,
      target.type,
      target.type === QueryElementTypes.Entity ? Handles.EntityLeft : Handles.RelationLeft,
      ''
    );

    // Add an edge to the graphology object
    const edgeId = this.addEdge(source.id, target.id, attributes);

    // Set the new nodes in the query builder slice TODO: maybe remove for efficiency
    store.dispatch(setQuerybuilderNodes(this.export()));

    return edgeId;
  }
}

/** Calculates the width and height of a query builder pill.
 * DEPENDS ON STYLING, if styling changed, change this.
 */
export function calcWidthHeightOfPill(attributes: Attributes): {
  width: number;
  height: number;
} {
  const { type, name } = attributes;

  let w = 0;
  let h = 0;
  switch (type) {
    case 'entity': {
      // calculate width and height of entity pill
      w = Math.min(name.length, 20) * widthPerFontsize[10]; // for fontsize 10px

      const widthOfPillWithoutText = 42.1164; // WARNING: depends on styling
      w += widthOfPillWithoutText;
      h = 20;
      break;
    }
    case 'relation': {
      // calculate width and height of relation pill
      w = Math.min(name.length, 20) * widthPerFontsize[10]; // for fontsize 10px

      const widthOfPillWithoutText = 56.0666; // WARNING: depends on styling
      w += widthOfPillWithoutText;
      h = 20;
      break;
    }
    case 'attribute': {
      // calculate width and height of relation pill
      const pixelsPerChar = widthPerFontsize[6]; // for fontsize 10px
      w = name.length * pixelsPerChar;

      const { datatype, operator } = attributes;
      let value = attributes['value'];
      if (!datatype || !operator) return { width: 0, height: 0 };
      if (!value) value = '?';

      // Add width of operator
      w += operator.length * widthPerFontsize[7];
      // use a max of 10, because max-width is set to 10ch;
      w += Math.min(value.length, 10) * widthPerFontsize[6];

      const widthOfPillWithoutText = 25.6666; // WARNING: depends on styling
      w += widthOfPillWithoutText;
      h = 12;
      break;
    }
  }

  return { width: w, height: h };
}

/** Interface for x and y position of node */
export interface NodePosition extends XYPosition {}

/** Returns from-position of relation node */
export function RelationPosToFromEntityPos(position: XYPosition): NodePosition {
  return { x: position.x - 60, y: position.y - 40 };
}

/** Returns to-position of relation node */
export function RelationPosToToEntityPos(position: XYPosition): NodePosition {
  return { x: position.x + 400, y: position.y };
}

/** Default position; x=0, y=0 */
export const DefaultPosition = { x: 0, y: 0 };
