import {
  setQuerybuilderNodes,
  useAppDispatch,
  useConfig,
  useQuerybuilderGraph,
  useQuerybuilderGraphology,
  useSchemaGraph,
  useSchemaGraphology,
} from '@graphpolaris/shared/lib/data-access/store';
import ReactFlow, {
  ReactFlowProvider,
  Background,
  Node,
  isNode,
  ReactFlowInstance,
  Controls,
  ControlButton,
  NodeChange,
  ConnectionMode,
  NodeMouseHandler,
  OnConnectStartParams,
  useReactFlow,
  Edge,
  Connection,
  HandleType,
  XYPosition,
  Position,
  OnEdgesChange,
  NodePositionChange,
} from 'reactflow';
import styles from './querybuilder.module.scss';

import React, { ReactComponentElement, useMemo, useRef, useEffect, useCallback, useState, DragEventHandler } from 'react';
import { AttributePill, ConnectionDragLine, ConnectionLine, EntityFlowElement, RelationPill } from '../pills';
import { dragPillStarted, dragPillStopped, movePillTo } from '../pills/dragging/dragPill';
import { Settings as SettingsIcon, Delete as DeleteIcon, ImportExport as ExportIcon } from '@mui/icons-material';
import { clearQB } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import { RelationPosToFromEntityPos, RelationPosToToEntityPos } from '@graphpolaris/shared/lib/querybuilder/model/graphology/utils';
import { useDispatch } from 'react-redux';
import {
  Card,
  CardContent,
  Dialog,
  DialogTitle,
  Grid,
  List,
  ListItem,
  ListItemButton,
  ListItemText,
  Paper,
  PaperProps,
  Typography,
} from '@mui/material';
import { Handles, NodeAttribute, QueryElementTypes, QueryGraphNodes, createReactFlowElements, fromHandleId, getHandleId } from '../model';
import Draggable from 'react-draggable';
import { GeneralDescription, InputNodeType } from '../model/logic/general';
import {
  MathFilterTypes,
  MathFilters,
  MathFunctionTypes,
  StringFilterTypes,
  MathFunctions,
  StringFunctions,
  StringFilters,
  StringFunctionTypes,
  AllLogicDescriptions,
  AllLogicMap,
} from '../model/logic';
import LogicPill from '../pills/customFlowPills/logicpill/logicpill';
import { current } from '@reduxjs/toolkit';
import { SchemaAttributeTypes } from '../../schema';

const nodeTypes = {
  entity: EntityFlowElement,
  relation: RelationPill,
  logic: LogicPill,
  attribute: AttributePill,
};
const edgeTypes = {
  connection: ConnectionLine,
  attribute_connection: ConnectionLine,
};

/**
 * This is the main querybuilder component. It is responsible for holding all pills and fire off the visual part of the querybuilder panel logic
 */
export const QueryBuilderInner: React.FC = () => {
  const [openPopup, setOpenPopup] = useState(false);
  const reactFlowWrapper = useRef<HTMLDivElement>(null);

  const schema = useSchemaGraphology();
  const graphologyGraph = useQuerybuilderGraphology();
  const graph = useQuerybuilderGraph();
  const config = useConfig();
  const dispatch = useDispatch();
  const isDraggingPill = useRef(false);
  const connectingNodeId = useRef<{
    params: OnConnectStartParams;
    position: XYPosition;
    node: QueryGraphNodes;
    attribute: NodeAttribute;
  } | null>(null);
  const reactFlow = useReactFlow();
  const isEdgeUpdating = useRef(false);
  const isOnConnect = useRef(false);

  const onInit = (reactFlowInstance: ReactFlowInstance) => {
    setTimeout(() => reactFlow.fitView(), 0);
  };

  const elements = useMemo(() => createReactFlowElements(graphologyGraph), [graph]);

  /**
   * Called when a node is dragged in querybuilder to allow for movement
   * @param event
   * @param node
   */
  const onNodeDrag = (event: React.MouseEvent<Element, MouseEvent>, node: Node<any>, nodes: Node[]) => {
    // console.log(nodes);
    // // Get the node in the elements list to get the previous location
    // const pNode = elements.nodes.find((e) => e?.id === node?.id);
    // if (!(pNode && isNode(pNode))) return;
    // // This is then used to calculate the delta position
    // const dx = node.position.x - pNode.position.x;
    // const dy = node.position.y - pNode.position.y;
    // // Check if we started dragging, if so, call the drag started usecase
    // if (!isDraggingPill.current) {
    //   dragPillStarted(node.id, graphologyGraph);
    //   isDraggingPill.current = true;
    // }
    // // Call the drag usecase
    // movePillTo(node.id, graphologyGraph, dx, dy, node.position);
    // // Dispatch the new graphology object, so reactflow will get rerendered
    // dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  };

  /**
   * Called when a node is released from dragging in querybuilder
   * @param event
   * @param node
   **/
  const onNodeDragStop = (event: React.MouseEvent<Element, MouseEvent>, node: Node<any>) => {
    // isDraggingPill.current = false;
    //
    // // Call the drag pill stopped usecase
    // dragPillStopped(node.id, graphologyGraph);
    //
    // // Dispatch the new graphology object, so reactflow will get rerendered
    // dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  };

  /**
   * Clears all nodes in the graph.
   */
  function clearAllNodes() {
    dispatch(clearQB());
  }

  /**
   * Clears all nodes in the graph.
   *  TODO: only works if the node is clicked and not moved (maybe use onSelectionChange)
   */
  function onNodesDelete(nodes: Node[]) {
    nodes.forEach((n) => {
      graphologyGraph.dropNode(n.id);
    });

    // Dispatch the new graphology object, so reactflow will get re-rendered
    dispatch(setQuerybuilderNodes(graphologyGraph.export()));
  }

  /**
   * TODO?
   */
  function onNodesChange(nodes: NodeChange[]) {
    // console.log(nodes);
    nodes.forEach((n) => {
      // console.log(graphologyGraph.findNode((gn) => gn === n?.id));
      if (n.type !== 'position') {
        // updated = true;
        // graphologyGraph.updateAttributes(n.id, n.data);
      } else {
        const node = n as NodePositionChange;
        // Get the node in the elements list to get the previous location
        const pNode = elements.nodes.find((e) => e?.id === node?.id);
        if (!(pNode && isNode(pNode)) || !node?.position) return;
        // This is then used to calculate the delta position
        const dx = node.position.x - pNode.position.x;
        const dy = node.position.y - pNode.position.y;
        // Check if we started dragging, if so, call the drag started usecase
        if (!isDraggingPill.current) {
          dragPillStarted(node.id, graphologyGraph);
          isDraggingPill.current = true;
        }
        // Call the drag usecase
        movePillTo(node.id, graphologyGraph, dx, dy, node.position);
        // Dispatch the new graphology object, so reactflow will get rerendered
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
      }
    });
  }

  const onDragOver = (event: React.DragEvent<HTMLDivElement>): void => {
    event.preventDefault();
    event.dataTransfer.dropEffect = 'move';
  };

  /**
   * The onDrop is called when the user drops an element from the schema onto the QueryBuilder.
   * In the onDrop query elements will be created based on the data stored in the drag event (datastrasfer).
   * @param event Drag event.
   */
  const onDrop = (event: React.DragEvent<HTMLDivElement>): void => {
    event.preventDefault();

    // The dropped element should be a valid reactflow element
    const data: string = event.dataTransfer.getData('application/reactflow');
    if (data.length == 0 || !reactFlow) return;

    const dragData = JSON.parse(data);
    console.log(reactFlowWrapper);

    const bounds = reactFlowWrapper?.current?.getBoundingClientRect() || { x: 0, y: 0 };
    const position = reactFlow.project({
      //TODO: this position should be centre of entity, rather than topleft
      x: event.clientX - bounds.x,
      y: event.clientY - bounds.y,
    });

    switch (dragData.type) {
      case QueryElementTypes.Entity:
        graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Entity,
          x: position.x,
          y: position.y,
          name: dragData.name,
        });
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
        break;
      // Creates a relation element and will also create the 2 related entities together with the connections
      case QueryElementTypes.Relation:
        const relation = graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Relation,
          x: position.x,
          y: position.y,
          depth: { min: 0, max: 1 },
          // name: dragData.name,
          name: dragData.collection,
          collection: dragData.collection,
        });
        const leftEntity = graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Entity,
          ...RelationPosToFromEntityPos(position),
          name: dragData.from,
        });
        const rightEntity = graphologyGraph.addPill2Graphology({
          type: QueryElementTypes.Entity,
          ...RelationPosToToEntityPos(position),
          name: dragData.to,
        });

        graphologyGraph.addEdge2Graphology(leftEntity, relation, {
          type: 'connection',
        });
        graphologyGraph.addEdge2Graphology(relation, rightEntity, {
          type: 'connection',
        });

        if (config.autoSendQueries) {
          // sendQuery();
        }

        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
        break;
      // Creates an attribute element with the correct dataType
      // case QueryElementTypes.Attribute:
      //   createNodeFromSchema(
      //     QueryElementTypes.Attribute,
      //     position,
      //     dragData.name,
      //     dragData.datatype
      //   );
      //   break;
    }
  };

  const onNodeMouseEnter = (event: React.MouseEvent, node: Node) => {
    console.log(node);
    console.log(schema.getNodeAttribute(node.type, 'attributes'));
  };

  const onNodeMouseLeave = (event: React.MouseEvent, node: Node) => {
    console.log(node);
  };

  const onConnect = useCallback(
    (connection: Connection) => {
      if (!isEdgeUpdating.current) {
        isOnConnect.current = true;
        graphologyGraph.addEdge(connection.source, connection.target, {
          type: 'connection',
          sourceHandle: connection.sourceHandle,
          targetHandle: connection.targetHandle,
        });
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
      }
    },
    [graph]
  );

  const onConnectStart = useCallback(
    (event: React.MouseEvent | React.TouchEvent, params: OnConnectStartParams) => {
      if (!params?.handleId) return;

      let node = graphologyGraph.getNodeAttributes(params.nodeId);
      const { attributeName, attributeType } = fromHandleId(params.handleId);
      // console.log(attributeName, attributeType, node.attributes?.filter((a) => a.name === attributeName)?.[0]);

      connectingNodeId.current = {
        params,
        node,
        position: { x: 0, y: 0 },
        attribute: { name: attributeName, type: attributeType as SchemaAttributeTypes, handleId: params.handleId },
      };
    },
    [graph]
  );

  const onConnectEnd = useCallback(
    (event: any) => {
      const targetIsPane = event.target.classList.contains('react-flow__pane');
      if (isOnConnect.current) {
        isOnConnect.current = false;
        return;
      }

      if (targetIsPane && !isEdgeUpdating.current) {
        let clientX: number = 0;
        let clientY: number = 0;

        if ('touches' in event) clientX = event?.touches?.[0]?.clientX;
        else if ('clientX' in event) clientX = event?.clientX;
        if ('touches' in event) clientY = event?.touches?.[0]?.clientY;
        else if ('clientY' in event) clientY = event?.clientY;

        if (reactFlowWrapper.current) {
          const { top, left } = reactFlowWrapper.current.getBoundingClientRect();
          clientX -= left;
          clientY -= top;
        }
        const position = reactFlow.project({ x: clientX, y: clientY });
        if (connectingNodeId?.current) connectingNodeId.current.position = position;

        setOpenPopup(true);
      }
    },
    [reactFlow.project]
  );

  const onNewNodeFromPopup = (value: AllLogicDescriptions, type: InputNodeType) => {
    if (connectingNodeId.current === null || connectingNodeId.current?.params?.handleId == null) return;
    const params = connectingNodeId.current.params;
    const position = connectingNodeId.current.position;

    const logic = AllLogicMap[value.key];
    const firstLeftLogicInput = logic.inputs.filter((input) => input.position === Position.Left)?.[0];
    if (!firstLeftLogicInput) return;

    // logicAttributes[0].handles = [connectingNodeId.current.handleId];
    const logicNode = graphologyGraph.addPill2Graphology({
      name: value.name,
      type: QueryElementTypes.Logic,
      x: position.x,
      y: position.y,
      logic: logic,
    });
    if (!logicNode?.id) throw new Error('Logic node has no id');

    graphologyGraph.addEdge(params.nodeId, logicNode.id, {
      type: 'connection',
      sourceHandle: params.handleId, // newAttribute data?
      targetHandle: getHandleId(logicNode.id, logicNode.name, logicNode.type, firstLeftLogicInput.name, firstLeftLogicInput.type.join(''), {
        extra: 'left',
      }),
    });

    dispatch(setQuerybuilderNodes(graphologyGraph.export()));
    setOpenPopup(false);
  };

  const onEdgeUpdateStart = useCallback(() => {
    isEdgeUpdating.current = true;
  }, []);

  const onEdgeUpdate = useCallback(
    (oldEdge: Edge, newConnection: Connection) => {
      if (isEdgeUpdating.current) {
        isEdgeUpdating.current = false;

        if (graphologyGraph.hasEdge(oldEdge.id)) {
          graphologyGraph.dropEdge(oldEdge.id);
        }

        graphologyGraph.addEdge(newConnection.source, newConnection.target, {
          type: 'connection',
          sourceHandle: newConnection.sourceHandle,
          targetHandle: newConnection.targetHandle,
        });
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
      }
    },
    [graph]
  );

  const onEdgesChange = (params: OnEdgesChange) => {
    console.log('edges change', params);
  };

  const onEdgeUpdateEnd = useCallback(
    (event: MouseEvent | TouchEvent, edge: Edge, handleType: HandleType) => {
      if (isEdgeUpdating.current) {
        if (graphologyGraph.hasEdge(edge.id)) {
          graphologyGraph.dropEdge(edge.id);
        }
        dispatch(setQuerybuilderNodes(graphologyGraph.export()));
        // setEdges((eds) => eds.filter((e) => e.id !== edge.id));
      }
      isEdgeUpdating.current = false;
    },
    [graph]
  );

  return (
    <div ref={reactFlowWrapper} className={styles.full}>
      <PopupMenu
        open={openPopup}
        type={connectingNodeId.current?.attribute?.type || 'int'}
        onClose={() => {
          setOpenPopup(false);
        }}
        onClick={onNewNodeFromPopup}
      />
      <ReactFlow
        edges={elements.edges}
        nodes={elements.nodes}
        snapGrid={[10, 10]}
        snapToGrid
        nodeTypes={nodeTypes}
        edgeTypes={edgeTypes}
        connectionLineComponent={ConnectionDragLine}
        // connectionMode={ConnectionMode.Loose}
        onInit={onInit}
        onNodesChange={onNodesChange}
        onNodeDrag={onNodeDrag}
        onNodeDragStop={onNodeDragStop}
        onDragOver={onDragOver}
        onConnect={onConnect}
        onConnectStart={onConnectStart}
        onConnectEnd={onConnectEnd}
        // onNodeMouseEnter={onNodeMouseEnter}
        // onNodeMouseLeave={onNodeMouseLeave}
        onEdgeUpdate={onEdgeUpdate}
        onEdgeUpdateStart={onEdgeUpdateStart}
        onEdgeUpdateEnd={onEdgeUpdateEnd}
        onDrop={onDrop}
        // onNodesDelete={onNodesDelete}
        // onNodesChange={onNodesChange}
        deleteKeyCode="Backspace"
        className={styles.reactflow}
        attributionPosition="top-right"
      >
        <Background gap={10} size={0.7} />
        {elements && elements.edges.length == 0 && elements.nodes.length == 0 && (
          <Card variant="outlined" sx={{ minWidth: 275, marginTop: 3, marginRight: 10 }}>
            <CardContent>
              <Typography sx={{ fontSize: 20 }} color="text.secondary">
                Drag some node from the Schema panel
              </Typography>
            </CardContent>
          </Card>
        )}

        <Controls showZoom={false} showInteractive={false} className={styles.controls}>
          <ControlButton className={styles.buttons} title={'Remove all elements'} onClick={() => clearAllNodes()}>
            <DeleteIcon />
          </ControlButton>
          <ControlButton
            className={styles.buttons}
            title={'Export querybuilder'}
            onClick={(event) => {
              event.stopPropagation();
              // this.setState({
              //   ...this.state,
              //   exportMenuAnchor: event.currentTarget,
              // });
            }}
          >
            <ExportIcon />
          </ControlButton>
          <ControlButton
            className={styles.buttons}
            title={'Other settings'}
            onClick={(event) => {
              event.stopPropagation();
              // this.setState({
              //   ...this.state,
              //   settingsMenuAnchor: event.currentTarget,
              // });
            }}
          >
            <SettingsIcon />
          </ControlButton>
        </Controls>
      </ReactFlow>
    </div>
  );
};

// const math_functions = ['AVG', 'COUNT', 'MAX', 'MIN', 'SUM', 'ROUND', 'CEIL', 'FLOOR', '+', '-', '*', '/', '%', 'CUSTOM'];
// const string_functions = ['CONCAT', 'LOWER', 'UPPER', 'SUBSTRING', 'TRIM'];
// const math_filters = ['==', '!=', '>', '<', '>=', '<='];
// const string_filters = ['==', '!=', 'LIKE', 'NOT LIKE', 'IN', 'NOT IN'];

function PaperComponent(props: PaperProps) {
  return (
    <Draggable handle="#draggable-dialog-title" cancel={'[class*="MuiDialogContent-root"]'}>
      <Paper {...props} />
    </Draggable>
  );
}

function PopupMenu(props: {
  open: boolean;
  type: InputNodeType;
  onClose: () => void;
  onClick: (value: AllLogicDescriptions, type: InputNodeType) => void;
}) {
  const handleClose = () => {
    props.onClose();
  };

  const handleListItemClick = (value: AllLogicDescriptions, type: InputNodeType) => {
    props.onClick(value, type);
  };

  const generateList = (list: AllLogicDescriptions[], type: InputNodeType) => (
    <Grid container sx={{ pt: 0 }}>
      {list.map((f, i) => (
        <Grid item key={JSON.stringify(f) + type + i}>
          <ListItemButton onClick={() => handleListItemClick(f, type)} key={f + type}>
            {/* <ListItemAvatar>
                <Avatar sx={{ bgcolor: blue[100], color: blue[600] }}>
                  <PersonIcon />
                </Avatar>
              </ListItemAvatar> */}
            <ListItemText primary={f.name} secondary={f.description} />
          </ListItemButton>
        </Grid>
      ))}
    </Grid>
  );
  return (
    <Dialog onClose={handleClose} open={props.open} PaperComponent={PaperComponent}>
      <DialogTitle>Add New Node</DialogTitle>
      <DialogTitle>Collection Nodes</DialogTitle>
      {props.type === 'float' && generateList(MathFunctions, props.type)}
      {props.type === 'int' && generateList(MathFunctions, props.type)}
      {props.type === 'string' && generateList(StringFunctions, props.type)}
      <DialogTitle>Filter Nodes</DialogTitle>
      {props.type === 'float' && generateList(MathFilters, props.type)}
      {props.type === 'int' && generateList(MathFilters, props.type)}
      {props.type === 'string' && generateList(StringFilters, props.type)}
    </Dialog>
  );
}

export const QueryBuilderPills = (props: { onClose: () => void; onClick: (value: AllLogicDescriptions, type: InputNodeType) => void }) => {
  const onDragStart = (event: React.DragEvent, nodeType: InputNodeType) => {
    console.log('drag start', nodeType);

    event.dataTransfer.setData('application/reactflow', nodeType);
    event.dataTransfer.effectAllowed = 'move';
  };

  const generateList = (list: AllLogicDescriptions[], type: InputNodeType) => (
    <List>
      {list.map((f, i) => (
        <ListItem key={JSON.stringify(f) + type + i}>
          {/* <ListItemAvatar>
                <Avatar sx={{ bgcolor: blue[100], color: blue[600] }}>
                  <PersonIcon />
                </Avatar>
              </ListItemAvatar> */}

          <ListItemText
            draggable
            primary={f.name}
            secondary={f.description}
            onDragStart={(event) => onDragStart(event, type)}
            key={f + type}
          />
        </ListItem>
      ))}
    </List>
  );

  return (
    <aside>
      <div>{generateList(MathFunctions, 'string')}</div>
    </aside>
  );
};

export const QueryBuilder = () => {
  return (
    <div
      style={{
        width: '100%',
        height: '100%',
        display: 'flex',
        gap: '1rem',
      }}
    >
      {/* <QueryBuilderPills /> */}
      <ReactFlowProvider>
        <QueryBuilderInner />
      </ReactFlowProvider>
    </div>
  );
};

// export const QueryBuilder = () => {
//   const graphologyGraph = useQuerybuilderGraphology();
//   const schema = useSchemaGraphology();
//   const graph = useQuerybuilderGraph();
//   const config = useConfig();
//   const dispatch = useDispatch();

//   return (
//     <QueryBuilderRaw
//       graphologyGraph={graphologyGraph}
//       schema={schema}
//       graph={graph}
//       config={config}
//     />
//   );
// };

export default QueryBuilder;
