/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

/* istanbul ignore file */
/* The comment above was added so the code coverage wouldn't count this file towards code coverage.
 * We do not test components/renderfunctions/styling files.
 * See testing plan for more details.*/
import {
  Autocomplete,
  AutocompleteRenderGroupParams,
  AutocompleteRenderInputParams,
  ListSubheader,
  TextField,
  Typography,
  makeStyles,
  useMediaQuery,
  useTheme,
} from '@mui/material';
import React from 'react';
import { VariableSizeList, ListChildComponentProps } from 'react-window';
import styles from './OptimizedAutocomplete.module.scss';

const LISTBOX_PADDING = 8; // px

function renderRow(props: ListChildComponentProps) {
  const { data, index, style } = props;
  return React.cloneElement(data[index], {
    style: {
      ...style,
      top: (style.top as number) + LISTBOX_PADDING,
    },
  });
}

const OuterElementContext = React.createContext({});

// eslint-disable-next-line react/display-name
const OuterElementType = React.forwardRef<HTMLDivElement>((props, ref) => {
  const outerProps = React.useContext(OuterElementContext);
  return <div ref={ref} {...props} {...outerProps} />;
});

function useResetCache(data: any) {
  const ref = React.useRef<VariableSizeList>(null);
  React.useEffect(() => {
    if (ref.current != null) {
      ref.current.resetAfterIndex(0, true);
    }
  }, [data]);
  return ref;
}

// Adapter for react-window
const ListboxComponent = React.forwardRef<HTMLDivElement>(function ListboxComponent(props: any, ref) {
  const { children, ...other } = props;
  const itemData = React.Children.toArray(children);
  const theme = useTheme();
  const smUp = useMediaQuery(theme.breakpoints.up('sm'), { noSsr: true });
  const itemCount = itemData.length;
  const itemSize = smUp ? 36 : 48;

  const getChildSize = (child: React.ReactNode) => {
    if (React.isValidElement(child) && child.type === ListSubheader) {
      return 48;
    }

    return itemSize;
  };

  const getHeight = () => {
    if (itemCount > 8) {
      return 8 * itemSize;
    }
    return itemData.map(getChildSize).reduce((a, b) => a + b, 0);
  };

  const gridRef = useResetCache(itemCount);

  return (
    <div ref={ref}>
      <OuterElementContext.Provider value={other}>
        <VariableSizeList
          itemData={itemData}
          height={getHeight() + 2 * LISTBOX_PADDING}
          width="100%"
          ref={gridRef}
          outerElementType={OuterElementType}
          innerElementType="ul"
          itemSize={(index: number) => getChildSize(itemData[index])}
          overscanCount={5}
          itemCount={itemCount}
          itemKey={(index: number) => 'autolist-key' + index}
        >
          {renderRow}
        </VariableSizeList>
      </OuterElementContext.Provider>
    </div>
  );
});

const renderGroup = (params: AutocompleteRenderGroupParams) => [
  <ListSubheader key={params.key} component="div">
    {params.group}
  </ListSubheader>,
  params.children,
];

type OptimizedAutocomplete = {
  currentValue: string;
  options: string[];
  useMaterialStyle?: { label: string; helperText: string };
  /** Called when the value of the input field changes. */
  onChange?(value: string): void;
  /** Called when the user leaves focus of the input field. */
  onLeave?(value: string): void;
};
/** Renders the autocomplete input field with the given props. */
export default function OptimizedAutocomplete(props: OptimizedAutocomplete) {
  let newValue = props.currentValue;
  console.log(props);

  return (
    <Autocomplete
      id="optimized-autocomplete"
      style={{ width: 200 }}
      disableListWrap
      autoHighlight
      className={styles.listbox}
      ListboxComponent={ListboxComponent as React.ComponentType<React.HTMLAttributes<HTMLElement>>}
      renderGroup={renderGroup}
      options={props.options}
      // groupBy={(option: string) => option[0].toUpperCase()}
      renderInput={(params: AutocompleteRenderInputParams) =>
        props.useMaterialStyle != undefined ? (
          <TextField
            {...params}
            variant="standard"
            // size="small"
            label={props.useMaterialStyle.label}
            helperText={props.useMaterialStyle.helperText}
          />
        ) : (
          <div ref={params.InputProps.ref}>
            <input style={{ width: 200 }} type="text" {...params.inputProps} autoFocus />
          </div>
        )
      }
      // defaultValue={props.currentValue}
      placeholder={props.currentValue}
      // renderOption={(props, option, state) => <Typography noWrap>{option}</Typography>}
      onChange={(_: React.ChangeEvent<any>, value: string | null) => {
        console.log(value);

        newValue = value || '?';
        if (props.onChange) props.onChange(newValue);
      }}
      onBlur={() => {
        if (props.onLeave) props.onLeave(newValue);
      }}
      onKeyDown={(e: React.KeyboardEvent<HTMLInputElement>): void => {
        if (e.key == 'Enter' && props.onLeave) props.onLeave(newValue);
      }}
    />
  );
}
